import ipaddress
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Optional, Union
from dataclasses import dataclass
from warnings import warn

import whodap
from tldextract.tldextract import TLDExtract

from .client import (
    ASNClient,
    DomainClient,
    NumberClient,
    convert_to_ip,
)
from .errors import NotFoundError, GeneralError, QueryError, WhoIsError

__all__ = [
    "aio_rdap",
    "aio_whois",
    "whois",
    "rdap",
    "aio_whois_domain",
    "aio_whois_ipv4",
    "aio_whois_ipv6",
    "aio_rdap_domain",
    "aio_rdap_ipv4",
    "aio_rdap_ipv6",
    "aio_rdap_asn",
    "rdap_domain",
    "rdap_ipv4",
    "rdap_ipv6",
    "rdap_asn",
    "whois_domain",
    "whois_ipv4",
    "whois_ipv6",
    "ASNClient",
    "DomainClient",
    "NumberClient",
    "NotFoundError",
    "WhoIsError",
    "GeneralError",
    "QueryError",
]
__version__ = "1.1.9"


def whois(
    search_term: Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,  # todo: deprecate and remove this argument
    find_authoritative_server: bool = True,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> tuple[str, dict]:
    """
    Performs a WHOIS query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: DEPRECATED - If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param find_authoritative_server: This parameter only applies to domain queries. If True (default), asyncwhois
        will attempt to find the authoritative response, otherwise if False, asyncwhois will only query the whois server
        associated with the given TLD as specified in the IANA root db (`asyncwhois/servers/domains.py`).
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the query output
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return NumberClient(
            authoritative_only=authoritative_only,
            proxy_url=proxy_url,
            timeout=timeout,
        ).whois(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return NumberClient(
                authoritative_only=authoritative_only,
                proxy_url=proxy_url,
                timeout=timeout,
            ).whois(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return DomainClient(
                authoritative_only=authoritative_only,
                find_authoritative_server=find_authoritative_server,
                ignore_not_found=ignore_not_found,
                proxy_url=proxy_url,
                timeout=timeout,
                tldextract_obj=tldextract_obj,
            ).whois(search_term)
    else:
        return "", {}


def rdap(
    search_term: Union[int, str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,
    whodap_client: Union[
        whodap.DNSClient, whodap.IPv4Client, whodap.IPv6Client, whodap.ASNClient
    ] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> tuple[str, dict]:
    """
    Performs an RDAP query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param whodap_client: An optional preconfigured instance of either an ASN, DNS, IPv4, or IPv6 `whodap` client
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return NumberClient(
            authoritative_only=authoritative_only,
            whodap_client=whodap_client,
        ).rdap(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return NumberClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
            ).rdap(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return DomainClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
                tldextract_obj=tldextract_obj,
            ).rdap(search_term)
    elif isinstance(search_term, int):
        return ASNClient(whodap_client=whodap_client).rdap(search_term)
    else:
        return "", {}


async def aio_whois(
    search_term: str,
    authoritative_only: bool = False,  # todo: deprecate and remove this argument
    find_authoritative_server: bool = True,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> tuple[str, dict]:
    """
    Performs a WHOIS query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: DEPRECATED - If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param find_authoritative_server: This parameter only applies to domain queries. If True (default), asyncwhois
        will attempt to find the authoritative response, otherwise if False, asyncwhois will only query the whois server
        associated with the given TLD as specified in the IANA root db (`asyncwhois/servers/domains.py`).
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the query output
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return await NumberClient(
            authoritative_only=authoritative_only,
            proxy_url=proxy_url,
            timeout=timeout,
        ).aio_whois(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return await NumberClient(
                authoritative_only=authoritative_only,
                proxy_url=proxy_url,
                timeout=timeout,
            ).aio_whois(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return await DomainClient(
                authoritative_only=authoritative_only,
                ignore_not_found=ignore_not_found,
                proxy_url=proxy_url,
                timeout=timeout,
                tldextract_obj=tldextract_obj,
                find_authoritative_server=find_authoritative_server,
            ).aio_whois(search_term)
    else:
        return "", {}


async def aio_rdap(
    search_term: Union[int, str, ipaddress.IPv4Address, ipaddress.IPv6Address],
    authoritative_only: bool = False,
    whodap_client: Union[
        whodap.DNSClient, whodap.IPv4Client, whodap.IPv6Client, whodap.ASNClient
    ] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> tuple[str, dict]:
    """
    Performs an RDAP query for the given `search_term`. If `search_term` is or can be cast to an
    instance of `ipaddress.IPv4Address` or `ipaddress.IPv6Address` then an IP search is performed
    otherwise a DNS search is performed.

    :param search_term: Any domain, URL, IPv4, or IPv6
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain,
        otherwise if True, only the authoritative response is returned.
    :param whodap_client: An optional preconfigured instance of either an ASN, DNS, IPv4, or IPv6 `whodap` client
    :param tldextract_obj: An optional preconfigured instance of `tldextract.TLDExtract` (used for parsing URLs)
    :returns: a tuple containing the WHOIS query text and a dictionary of key values parsed from the text
    """
    if isinstance(search_term, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return await NumberClient(
            authoritative_only=authoritative_only,
            whodap_client=whodap_client,
        ).aio_rdap(search_term)
    elif isinstance(search_term, str):
        try:
            search_term = convert_to_ip(search_term)
            return await NumberClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
            ).aio_rdap(search_term)
        except (ipaddress.AddressValueError, ValueError):
            return await DomainClient(
                authoritative_only=authoritative_only,
                whodap_client=whodap_client,
                tldextract_obj=tldextract_obj,
            ).aio_rdap(search_term)
    elif isinstance(search_term, int):
        return await ASNClient(whodap_client=whodap_client).aio_rdap(search_term)
    else:
        return "", {}


# ====================
# TODO: ALL the code below will be removed in a future release; it is here for backwards compatibility only


@dataclass
class DomainLookup:
    query_output: str
    parser_output: dict


@dataclass
class NumberLookup:
    query_output: str
    parser_output: dict


@dataclass
class ASNLookup:
    query_output: str
    parser_output: dict


def whois_domain(
    domain: str,
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> DomainLookup:
    """
    Performs domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the `query_output`
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.whois_domain` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = DomainClient(
        authoritative_only=authoritative_only,
        ignore_not_found=ignore_not_found,
        proxy_url=proxy_url,
        timeout=timeout,
        tldextract_obj=tldextract_obj,
    ).whois(domain)
    return DomainLookup(query_output, parser_output)


async def aio_whois_domain(
    domain: str,
    authoritative_only: bool = False,
    ignore_not_found: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
    tldextract_obj: TLDExtract = None,
) -> DomainLookup:
    """
    Performs asynchronous domain lookups with WHOIS.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param authoritative_only: If False (default), asyncwhois returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param ignore_not_found:  If False (default), the `NotFoundError` exception is raised if the `query_output`
        contains "no such domain" language. If True, asyncwhois will not raise `NotFoundError` exceptions.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url (e.g. 'socks5://host:port')
    :param timeout: Connection timeout. Default is 10 seconds.
    :param tldextract_obj: An optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.aio_whois_domain` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await DomainClient(
        authoritative_only=authoritative_only,
        ignore_not_found=ignore_not_found,
        proxy_url=proxy_url,
        timeout=timeout,
        tldextract_obj=tldextract_obj,
    ).aio_whois(domain)
    return DomainLookup(query_output, parser_output)


def rdap_domain(
    domain: str,
    httpx_client: Optional[Any] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> DomainLookup:
    """
    Performs an RDAP query for the given domain.
    Finds the authoritative RDAP server and parses the response from that server.

    :param domain: Any domain name or URL
        (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured instance of `httpx.AsyncClient`
    :param tldextract_obj: Optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.rdap_domain` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.DNSClient.new_client(httpx_client=httpx_client)
    query_output, parser_output = DomainClient(
        whodap_client=whodap_client,
        tldextract_obj=tldextract_obj,
    ).rdap(domain)
    return DomainLookup(query_output, parser_output)


async def aio_rdap_domain(
    domain: str,
    httpx_client: Optional[Any] = None,
    tldextract_obj: Optional[TLDExtract] = None,
) -> DomainLookup:
    """
    Performs an async RDAP query for the given domain name.

    :param domain: Any domain or URL (e.g. 'wikipedia.org' or 'https://en.wikipedia.org/wiki/WHOIS')
    :param httpx_client: Optional preconfigured instance of `httpx.AsyncClient`
    :param tldextract_obj: Optional preconfigured instance of `tldextract.tldextract.TLDExtract`
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.aio_rdap_domain` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.DNSClient.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = await DomainClient(
        whodap_client=whodap_client,
        tldextract_obj=tldextract_obj,
    ).aio_rdap(domain)
    return DomainLookup(query_output, parser_output)


def whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of DomainLookup
    """
    warn(
        "`asyncwhois.whois_ipv4` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).whois(ipv4)
    return NumberLookup(query_output, parser_output)


async def aio_whois_ipv4(
    ipv4: Union[IPv4Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv4 address.
    Finds the authoritative WHOIS server and parses the response from the server.

    :param ipv4: ip address as a str or `ipaddress.IPv4Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_whois_ipv4` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).aio_whois(ipv4)
    return NumberLookup(query_output, parser_output)


def rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv4 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.rdap_ipv4` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv4Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).rdap(ipv4)
    return NumberLookup(query_output, parser_output)


async def aio_rdap_ipv4(
    ipv4: Union[IPv4Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv4: IP address as a string or `ipaddress.IPv4Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_rdap_ipv4` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv4Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).aio_rdap(ipv4)
    return NumberLookup(query_output, parser_output)


def whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs a WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.whois_ipv6` is deprecated. Please use `asyncwhois.whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).whois(ipv6)
    return NumberLookup(query_output, parser_output)


async def aio_whois_ipv6(
    ipv6: Union[IPv6Address, str],
    authoritative_only: bool = False,
    proxy_url: str = None,
    timeout: int = 10,
) -> NumberLookup:
    """
    Performs an async WHOIS query for the given IPv6 address.
    Looks up the WHOIS server, submits the query, and then parses the response from the server.

    :param ipv6: ip address as a str or `ipaddress.IPv6Address` object
    :param authoritative_only: If False, returns the entire WHOIS query chain
        in `query_output`; If True only the authoritative response is included.
    :param proxy_url: Optional SOCKS4 or SOCKS5 proxy url
    :param timeout: Connection timeout. Default is 10 seconds.
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_whois_ipv6` is deprecated. Please use `asyncwhois.aio_whois` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    query_output, parser_output = await NumberClient(
        authoritative_only=authoritative_only,
        proxy_url=proxy_url,
        timeout=timeout,
    ).aio_whois(ipv6)
    return NumberLookup(query_output, parser_output)


def rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.rdap_ipv6` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).rdap(ipv6)
    return NumberLookup(query_output, parser_output)


async def aio_rdap_ipv6(
    ipv6: Union[IPv6Address, str], httpx_client: Optional[Any] = None
) -> NumberLookup:
    """
    Performs an async RDAP query for the given IPv6 address.

    :param ipv6: IP address as a string or `ipaddress.IPv6Address` object
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of NumberLookup
    """
    warn(
        "`asyncwhois.aio_rdap_ipv6` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = NumberClient(
        whodap_client=whodap_client,
    ).aio_rdap(ipv6)
    return NumberLookup(query_output, parser_output)


def rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.Client`
    :return: instance of ASNLookup
    """
    warn(
        "`asyncwhois.rdap_asn` is deprecated. Please use `asyncwhois.rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_client(httpx_client=httpx_client)
    query_output, parser_output = ASNClient(
        whodap_client=whodap_client,
    ).rdap(asn)
    return ASNLookup(query_output, parser_output)


async def aio_rdap_asn(asn: int, httpx_client: Optional[Any] = None) -> ASNLookup:
    """
    Performs an async RDAP query for the given Autonomous System Number.

    :param asn: The ASN number as an integer
    :param httpx_client: Optional preconfigured `httpx.AsyncClient`
    :return: instance of ASNLookup
    """
    warn(
        "`asyncwhois.aio_rdap_asn` is deprecated. Please use `asyncwhois.aio_rdap` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    whodap_client = None
    if httpx_client is not None:
        whodap_client = whodap.IPv6Client.new_aio_client(httpx_client=httpx_client)
    query_output, parser_output = ASNClient(
        whodap_client=whodap_client,
    ).aio_rdap(asn)
    return ASNLookup(query_output, parser_output)



from pprint import pprint

import asyncwhois
import whodap
import httpx


def main():
    domain = "axe.pizza"

    # WHOIS
    client = asyncwhois.DomainClient(ignore_not_found=True)

    query_string, parsed_dict = client.whois(domain)
    pprint(query_string)
    pprint(parsed_dict)

    # RDAP

    # Simple client example:
    client = asyncwhois.DomainClient()
    query_string, parsed_dict = client.rdap(domain)
    pprint(query_string)
    pprint(parsed_dict)

    # Fully configurable client example:
    # Proxy with `httpx`
    whodap_client = whodap.DNSClient.new_client(
        httpx_client=httpx.Client(proxies="https://proxy:8080")
    )
    client = asyncwhois.DomainClient(whodap_client=whodap_client)
    query_output, parser_output = client.rdap(domain)

    # **EXTERNAL DEPENDENCY** for SOCKS Proxies (`httpx_socks`)
    from httpx_socks import SyncProxyTransport

    transport = SyncProxyTransport.from_url("socks5://localhost:9050")
    whodap_client = whodap.DNSClient.new_client(
        httpx_client=httpx.Client(transport=transport)
    )
    client = asyncwhois.DomainClient(whodap_client=whodap_client)
    query_string, parsed_dict = client.rdap(domain)
    return


if __name__ == "__main__":
    main()



import textwrap
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone

import httpx
import pandas as pd
import stamina
from bs4 import BeautifulSoup as bs


iana_root_db_uri = "https://www.iana.org/domains/root/db"


@stamina.retry(
    on=httpx.HTTPError,
    attempts=3,
    wait_initial=2.0,
    wait_jitter=2.0,
    wait_max=10.0,
)
def get_html(url: str) -> str:
    resp = httpx.get(url)
    resp.raise_for_status()
    return resp.text


def parse_root_db_table(root_db_html: str) -> list[tuple[str, str, str, str]]:
    tlds = []
    soup = bs(root_db_html, "html.parser")
    table = soup.find("table", attrs={"id": "tld-table"})
    table_body = table.find("tbody")
    rows = table_body.find_all("tr")
    for row in rows:
        cols = row.find_all("td")
        # tld and type
        tld, tld_type, _ = [v.get_text() for v in cols]
        # parse the "href" where server info is
        tld_link = cols[0].find("a")
        if tld_link is not None:
            href = f'https://www.iana.org{tld_link.get("href")}'
            # get the "urlsafe" tld
            tld_safe = href.split("/")[-1].replace(".html", "")
            tlds.append((tld.lstrip("\n."), tld_safe, href, tld_type))

    return tlds


def find_server_on_page(tld_html: str) -> str | None:
    soup = bs(tld_html, "html.parser")
    text = soup.get_text()
    start = text.find("WHOIS Server:")
    if start >= 0:
        end = text[start:].find("\n")
        whois_server_substring = text[start : start + end]
        whois_server = whois_server_substring.split(": ")[-1]
        return whois_server
    return None


def get_server(link: str):
    html = get_html(link)
    server = find_server_on_page(html)
    return server


def create_module_file(df: pd.DataFrame) -> str:

    def enc(x):
        return x.replace("-", "_").upper()

    cc_tlds_df = df.loc[df.tld_type == "country-code", :]
    country_code_tlds = ""
    for i, row in cc_tlds_df.iterrows():
        country_code_tlds += f"{enc(row.punycode)} = "
        if row.server_name is None:
            country_code_tlds += "None\n\t"
        else:
            country_code_tlds += f'"{row.server_name}"\n\t'

    g_tlds_df = df.loc[df.tld_type == "generic", :]
    generic_tlds = ""
    for i, row in g_tlds_df.iterrows():
        generic_tlds += f"{enc(row.punycode)} = "
        if row.server_name is None:
            generic_tlds += "None\n\t"
        else:
            generic_tlds += f'"{row.server_name}"\n\t'

    s_tlds_df = df.loc[df.tld_type == "sponsored", :]
    sponsored_tlds = ""
    for i, row in s_tlds_df.iterrows():
        sponsored_tlds += f"{enc(row.punycode)} = "
        if row.server_name is None:
            sponsored_tlds += "None\n\t"
        else:
            sponsored_tlds += f'"{row.server_name}"\n\t'

    ts = datetime.now(tz=timezone.utc).strftime("%Y:%m:%d %H:%M:%S")

    return f"""
# Extracted from the IANA Root Zone Database: https://www.iana.org/domains/root/db
# on {ts} UTC by https://github.com/pogzyb/asyncwhois/tools/tld-server-updates.py

class CountryCodeTLD: 
\t{country_code_tlds}

class GenericTLD: 
\t{generic_tlds}

class SponsoredTLD: 
\t{sponsored_tlds}
"""


def main():
    # iana_root_db_html = get_html(iana_root_db_uri)
    # tld_table_data = parse_root_db_table(iana_root_db_html)
    #
    # tld_server_info = []
    # with ThreadPoolExecutor(max_workers=2) as pool:
    #     server_parsers = {}
    #     for tld, tld_safe, href, tld_type in tld_table_data:
    #         server_parsers[pool.submit(get_server, href)] = (
    #             tld,
    #             tld_safe,
    #             tld_type,
    #             href,
    #         )
    #
    #     for server_parser in as_completed(server_parsers):
    #         tld, tld_safe, tld_type, href = server_parsers[server_parser]
    #         server_name = server_parser.result()
    #         tld_server_info.append((tld, tld_safe, tld_type, server_name, href))
    #
    # df = pd.DataFrame.from_records(
    #     tld_server_info, columns=["tld", "punycode", "tld_type", "server_name", "href"]
    # )
    # df.to_csv("tld_servers.csv", index=False)

    # for debugging:
    df = pd.read_csv("tld_servers.csv")
    df["server_name"] = df["server_name"].apply(lambda x: None if pd.isnull(x) else x)

    file_content = create_module_file(df)
    with open("domains.py", "w") as out:
        out.write(file_content)


if __name__ == "__main__":
    main()



from typing import Any

from .parse import TLDBaseKeys
from .errors import NotFoundError
from . import tldparsers


class DomainParser:
    _no_match_checks = [
        "no match",
        "not found",
        "no entries found",
        "invalid query",
        "domain name not known",
        "no object found",
        "available for re-registration",
        "object does not exist",
        "domain you requested is not known",
    ]

    def __init__(self, ignore_not_found: bool = False) -> None:
        self.ignore_not_found = ignore_not_found

    def parse(self, blob: str, tld: str) -> dict[TLDBaseKeys, Any]:
        low_blob = blob.lower()
        if not self.ignore_not_found and any(
            n in low_blob for n in self._no_match_checks
        ):
            raise NotFoundError("Domain not found!")
        parser = self._init_parser(tld)
        return parser.parse(blob)

    @staticmethod
    def _init_parser(tld: str) -> tldparsers.TLDParser:
        """
        Retrieves the parser instance which can most accurately extract
        key/value pairs from the whois server output for the given `tld`.

        :param tld: the top level domain
        :return: instance of TLDParser or a TLDParser subclass
        """

        try:
            cls_ = getattr(tldparsers, f"Regex{tld.upper()}")
        except AttributeError:
            # The TLDParser can handle all "Generic" and some "Country-Code"
            # TLDs.  If the parsed output of lookup is not what you expect or
            # even incorrect, check and modify the existing Regex subclass or
            # create a new one.
            cls_ = tldparsers.TLDParser
        return cls_()



import asyncio
import sys
import unittest.mock as mock

import asyncwhois
import pytest


test_domain_name = "amazon.com"
mock_response = ("Domain Name: amazon.com", {"domain_name": test_domain_name})


if sys.version_info < (3, 8):

    @pytest.fixture()
    def mock_aio_whois_domain(mocker):
        future = asyncio.Future()
        future.set_result(mock_response)
        mocker.patch("asyncwhois.client.DomainClient.aio_whois", return_value=future)
        return future

else:

    @pytest.fixture()
    def mock_aio_whois_domain(mocker):
        async_mock = mock.AsyncMock(return_value=mock_response)
        mocker.patch("asyncwhois.client.DomainClient.aio_whois", side_effect=async_mock)
        return async_mock


@pytest.fixture()
def mock_whois_domain(mocker):
    mocker.patch(
        "asyncwhois.client.DomainClient.whois",
        return_value=mock_response,
    )


@pytest.mark.asyncio
async def test_aio_whois(mock_aio_whois_domain):
    q, p = await asyncwhois.aio_whois(test_domain_name)
    assert (
        f"domain name: {test_domain_name}" in q.lower()
    ), f"domain name: {test_domain_name} not in {q.lower()}"
    assert p.get("domain_name").lower() == test_domain_name


def test_whois(mock_whois_domain):
    q, p = asyncwhois.whois(test_domain_name)
    assert (
        f"domain name: {test_domain_name}" in q.lower()
    ), f"domain name: {test_domain_name} not in {q.lower()}"
    assert p.get("domain_name").lower() == test_domain_name
